// File_Avc - Info for AVC Video files
// Copyright (C) 2006-2007 Jerome Martinez, Zen@MediaArea.net
//
// This library is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this library. If not, see <http://www.gnu.org/licenses/>.
//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//---------------------------------------------------------------------------
#ifndef MediaInfo_AvcH
#define MediaInfo_AvcH
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#include "MediaInfo/File__Analyze.h"
//---------------------------------------------------------------------------

namespace MediaInfoLib
{

//***************************************************************************
// Class File_Avc
//***************************************************************************

class File_Avc : public File__Analyze
{
public :
    //In
    size_t Frame_Count_Valid;
    bool   FrameIsAlwaysComplete;

protected :
    //Information
    void HowTo (stream_t StreamKind);

public :
    File_Avc();

private :
    //Replacement of File__Base
    const int8u* Buffer_ToSave;
    size_t Buffer_Size_ToSave;

    //Buffer
    bool Header_Begin();
    void Header_Parse();
    bool Header_Parse_Fill_Size();
    bool Data_Parse();

    //Packets
    void slice_layer_without_partitioning_IDR();
    void slice_layer_without_partitioning_non_IDR();
    void slice_header();
    void slice_header_Fill();
    void seq_parameter_set();
    void pic_parameter_set();
    void sei();
    void sei_user_data_unregistered(int32u payloadSize);
    void sei_buffering_period(int32u payloadSize);
    void sei_pic_timing(int32u payloadSize);
    void sei_user_data_unregistered_x264(int32u payloadSize);
    void access_unit_delimiter();

    //Packets - SubElements
    void scaling_list(int32u ScalingList_Size);
    void vui_parameters();

    //Count of a Packets
    size_t Frame_Count;
    int32u frame_num_LastOne;

    //From seq_parameter_set
    int32u pic_width_in_mbs_minus1;
    int32u pic_height_in_map_units_minus1;
    int32u log2_max_frame_num_minus4;
    int8u  profile_idc;
    int8u  level_idc;
    int8u  aspect_ratio_idc;
    int16u sar_width;
    int16u sar_height;
    bool   frame_mbs_only_flag;

    //PS
    struct avc_stream
    {
        bool   Searching_Payload;

        avc_stream()
        {
            Searching_Payload=false;
        }
    };
    std::map<int8u, avc_stream> Stream;

    //Helpers
    bool Synchronize();
    bool Header_Parser_QuickSearch();
    bool Detect_NonAVC();
};

} //NameSpace

#endif
